/*
 Copyright 2017 JetBrains s.r.o.

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 */

/**
 * Module containing a straightforward cache implementation.
 *
 * @module @jetbrains/youtrack-scripting-api/cache
 */

var Cache = function (size, name) {
  if (typeof size !== 'number' || size < 1) {
    throw 'Size should be positive number';
  }
  this.maxSize = size;
  this.name = name;
  this.clear();
};

Cache.prototype.clear = function () {
  this.first = null;
  this.last = null;
  this.map = {}; // stores {key: key, value: cached_value, prev:prev_stored_item, next: next_stored_item} under key
  this.size = 0;
};

Cache.prototype.get = function (key, creator) {
  if (!key) {
    throw "Tried to retrieve a value by null key";
  }
  if (typeof key !== 'string') {
    throw "Key is not a string";
  }
  var found = this.map[key];
  if (found) {
    console.trace(this.name + ' cache hit, key: ' + key);
    if (this.first === found) {
      return found.value;
    }
    if (found.prev) {
      found.prev.next = found.next;
    }
    if (found.next) {
      found.next.prev = found.prev;
    } else {
      // it's the last. found.prev != null because otherwise the very first check in this branch would have led to return;
      this.last = found.prev;
    }
    found.prev = null;
    found.next = this.first;
    this.first = found;
    return found.value;
  } else {
    console.trace(this.name + ' cache missed, key: ' + key);
    var created = creator(key); // seem to be ok to allow caching nulls
    var newElem = {key: key, prev: null, value: created, next: this.first};
    this.map[key] = newElem;
    this.first = newElem;
    if (!this.last) {
      this.last = newElem;
    }
    if (this.size >= this.maxSize) {
      console.trace('Removing key ' + this.last.key + ' from cache ' + this.name);
      delete this.map[this.last.key];
      this.last = this.last.prev;
    } else {
      this.size++;
    }
    return created;
  }
};

exports.create = function (size, name) {
  return new Cache(size, name);
};