/*
 Copyright 2017 JetBrains s.r.o.

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 */

/**
 * Date and time related operations
 *
 * @module @jetbrains/youtrack-scripting-api/date-time
 */

var utils = Packages.jetbrains.youtrack.scripts.sandbox.Utils;
var Period = Packages.org.joda.time.Period;
var formatter = new Packages.org.joda.time.format.PeriodFormatterBuilder().rejectSignedValues(true).appendWeeks().appendSuffix('w').appendDays().appendSuffix('d').appendHours().appendSuffix('h').appendMinutes().appendSuffix('m').toFormatter();

/**
 * Parses string representation of a date. Use this method instead of Date.parse()
 * (see {@link https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Date/parse}
 * for explanation).
 *
 * @param {string} dateTimeString string representation of a date.
 *
 * @param {(string|string[])} [formats] a format that the dateTimeString matches or an array of formats.
 * In case an array is provided, the formats are tried sequentially unless the dateTimeString is successfully parsed.
 * If not specified, a current user's format is used. If there's no logged in user (e.g. when calling inside an on schedule rule),
 * system default format is used.
 *
 * @param {string} [timeZoneId] id of a timezone that the dateTimeString is in. Only effective in case a format that
 * matches the string does not provide any timestamp information. In case neither format(s) nor this parameter
 * provide any timezone information, a time zone of a current user will be taken. If there's no logged in user
 * (e.g. when calling inside an on schedule rule), system default format is used.
 * See {@link http://joda-time.sourceforge.net/timezones.html|JodaTime} for a list of timezone ids.
 *
 * @return {number} timestamp representation of the provided date.
 */
exports.parse = function (dateTimeString, formats, timeZoneId) {
  if (formats && !Array.isArray(formats)) {
    formats = [formats];
  }
  return utils.toUnixTimestamp(dateTimeString, formats || null, timeZoneId || null);
};

/**
 * Creates string representation of a timestamp.
 *
 * @param {number} timestamp timestamp to be formatted into a string.
 *
 * @param {string} [format] format of the output.
 *
 * @param {string} [timeZoneId] id of a timezone.
 *
 * @see {@link parse} for timezone id description
 *
 * @see {@link parse} for format string description.
 *
 * @return {string} string representation of the timestamp.
 */
exports.format = function (timestamp, format, timeZoneId) {
  return utils.formatTimestamp(timestamp, format || null, timeZoneId || null);
};

/**
 * Creates a period representation of an argument.
 *
 * @param {(number|string)} period duration in milliseconds as either a number or a string.
 * @return {Period} period representation of an argument
 * @see {@link PeriodProjectCustomField} for possible usages.
 */
exports.toPeriod = function (period) {
  if (typeof period === 'number') {
    return new Period(period);
  }
  var n = parseInt(period);
  if (!isNaN(n)) {
    return new Period(n);
  }
  return Period.parse(period, formatter);
};
