/*
 Copyright 2017 JetBrains s.r.o.

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 */

/**
 * Entities module.
 * @module @jetbrains/youtrack-scripting-api/entities
 */

/**
 @typedef {Object} Requirement
 @classdesc
 One element of {@link Requirements}
 @property {string} [name] optional field/entity name. If not provided,
 key for this requirement in {@link Requirements} object will be used.
 @property {string} [login] optional login, used for {@link User} requirements
 instead of name.
 @property {string} [id] optional id, used for {@link Issue} requirements instead of name.
 @property {boolean} [multi] optional flag, `false` by default. If `true`, a required field
 has to allow multiple values (if applicable).
 @property {string|Object} type field type. Can be one of the following: {@link Build}.fieldType,
 {@link OwnedField}.fieldType, {@link State}.fieldType, {@link EnumField}.fieldType,
 {@link ProjectVersion}.fieldType, {@link User}.fieldType, {@link UserGroup}.fieldType,
 {@link Field}.dateType, {@link Field}.floatType, {@link Field}.integerType,
 {@link Field}.stringType, {@link Field}.periodType, {@link User}, {@link UserGroup},
 {@link Project}, {@link Issue}, {@link IssueTag}, {@link SavedQuery}.
 */

/**
 @typedef {Object<string, Requirement>} Requirements

 @classdesc
 Requirements contain a set of entities that should be present on a given
 YouTrack instance for the rule to work as expected. Whenever one or more
 rule requirements are not met, corresponding errors are shown in
 workflow administration UI and the rule is not executed until all problems are fixed.
 This is a 'safety-net' function of requirements.

 A second function is a 'reference' one: each required object is plugged into
 the `context` object, so that you can reference entities from inside your
 context-dependent functions (like an `action` function).

 There are two types of requirements: project-wide and system-wide.
 Project-wide requirements contain a list of custom fields to be attached
 to each project which uses the rule as well as some values from fields value sets.
 System-wide requirements contain a list of other entities that should be present
 on a given instance: users and user groups, projects and issues, tags and saved searches.

 @example
 requirements: {
   P: {
     type: entities.EnumField.fieldType,
     name: 'Priority',
     M: { name: 'Major' },
     Normal: {}
   },

   ImportantPerson: {
     type: entities.User,
     login: 'superadmin'
   },
   OurTeam: {
     type: entities.UserGroup,
     name: 'integration-team'
   },
   Int: {
     type: entities.Project,
     name: 'Integration'
   },
   Ref: {
     type: entities.Issue,
     id: 'INT-483'
   },
   ToBeReleased: {
     type: entities.IssueTag,
     name: 'To be released'
   },
   Untested: {
     type: entities.SavedQuery,
     name: 'Not tested yet'
   }
 }
 */

/**
 * This function is called on certain events, depending on rule type:
 * when issue is changed (on-change rules), when command is executed (action rules),
 * or according to schedule (scheduled rules). This function is called separately
 * for each related issue.
 *
 * @example
 * action: function(ctx) {
 *   console.log('Action for issue ' + ctx.issue.id);
 * }
 *
 * @callback Issue~actionFunction
 *
 * @param {Object} ctx execution context. Along with the params mentioned below,
 * the context also contains objects that you describe in {@link Requirements}.
 * @param {Issue} ctx.issue current issue.
 * @param {User} ctx.currentUser a user by which this rule is executed.
 */

/**
 * This function is called to check if an action rule is applicable to an issue.
 *
 * @example
 * guard: function(ctx) {
 *   return ctx.issue.State.isResolved;
 * }
 *
 * @callback Issue~guardFunction
 *
 * @param {Object} ctx execution context. Along with the params mentioned below,
 * the context also contains objects that you describe in {@link Requirements}.
 * @param {Issue} ctx.issue current issue.
 * @param {User} ctx.currentUser a user by which this rule is executed.
 */


/**
 * @typedef {Object} Iterator
 *
 * @classdesc
 * Iterator allows traversing values of some collection one by one.
 */
/**
 * @method
 * @name next
 * @memberOf Iterator
 * @return {Object} an object with `done` and `value` properties:
 * `done` is `false` if there are elements left and `true` otherwise,
 * `value` is a next element of a collection if there is any
 * and `null` otherwise
 */

/**
 * @typedef {Object} Set
 *
 * @classdesc
 * The Set object stores unique values of any type, whether primitive values or
 * object references. The Set is used as a storage for all multiple values in
 * this API: multiple custom fields, links, issues in a project etc.
 * Values can be accessed directly (see first(), last(), get(index)),
 * via iterator (see entries(), values()), or traversed with forEach(visitor)
 * and find(predicate) methods.
 *
 * This Set implementation is based on
 * {@link https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Set}.
 * However, as we are using ECMAScript 5 now, our Set implementation just mimics ES 6 Set interface.
 * Please see details below.
 *
 * @property {number} [size] number of elements in a Set; please use accurately,
 * as for the big collections (like `project.issues`) size calculation may be
 * resource consuming.
 *
 * @property {Set} [added] elements that were added to a multi-valued field values during the current transaction.
 * Only relevant in case the Set represents a multi-valued property (field) of a persistent entity.
 *
 * @property {Set} [removed] elements that were removed from a multi-valued field values during the current transaction.
 * Only relevant in case the Set represents a multi-valued property (field) of a persistent entity.
 *
 * @property {boolean} [isChanged] `true` in case the Set represents a multi-valued property (field) of a persistent entity
 * and the field is changed in the current transaction.
 *
 * @returns {boolean} `true` in case the field is changed in the current transaction.
 *
 * @returns {Set} values that were added.
 */
/**
 * @method
 * @instance
 * @name get
 * @memberOf Set
 * @param {number} index
 * @returns {Object} an object at index position in a Set, or null
 * if Set have less elements than (index + 1)
 */
/**
 * @method
 * @instance
 * @name first
 * @memberOf Set
 * @returns {Object} first object in a collection (in an order of adding),
 * or null if collection is empty
 */
/**
 * @method
 * @instance
 * @name last
 * @memberOf Set
 * @returns {Object} last object in a collection (in an order of adding),
 * or null if collection is empty
 */
/**
 * Get an iterator over Set entries. The same as `values()`.
 * Use iterator when you need to traverse over entries until some condition
 * is met and modify them at the same time.
 *
 * @see forEach(visitor) if you need to modify all Set elements
 * @see find(predicate) if you need to find an element
 *
 * @example
 * // We want to find first Critical among issue subtasks and assign it.
 * var checkAndAssign = function(task) {
 *   if (task.fields.Priority.name === ctx.Priority.Critical.name) {
 *     task.fields.Assignee = ctx.currentUser;
 *     return true;
 *   }
 *   return false;
 * };
 * var iter = issue.links['parent for'].entries();
 * var v = iter.next();
 * while (!v.done && !checkAndAssign(v.value)) {
 *   v = iter.next();
 * }
 *
 * @method
 * @instance
 * @name entries
 * @memberOf Set
 * @returns {Iterator<Object>} iterator over collection values
 */
/**
 * Get an iterator over Set entries. The same as `entries()`.
 * @see entries() for details
 *
 * @method
 * @instance
 * @name values
 * @memberOf Set
 * @returns {Iterator<Object>} iterator over collection values
 */
/**
 * Apply a visitor function at each member of a collection.
 *
 * @example
 * issue.links['parent for'].forEach(function(child) {
 *   child.fields.Priority = issue.fields.Priority;
 * });
 *
 * @method
 * @instance
 * @name forEach
 * @memberOf Set
 * @param {function} visitor
 */
/**
 * Find the first element in a Set, for which predicate returns `true`.
 *
 * @example
 * issue.tags.find(function(tag) {
 *   tag.name === 'release';
 * });
 *
 * @method
 * @instance
 * @name find
 * @memberOf Set
 * @param {function} predicate a function with one argument that returns
 * either true or false for a given Set value
 * @returns {Object} first value found or null
 */
/**
 * @method
 * @instance
 * @name has
 * @memberOf Set
 * @param {Object} element
 * @returns {boolean} `true` if element is in Set, `false` otherwise
 */
/**
 * @method
 * @instance
 * @name isEmpty
 * @memberOf Set
 * @returns {boolean} `true` if Set contains no elements, `false` otherwise
 */
/**
 * @method
 * @instance
 * @name isNotEmpty
 * @memberOf Set
 * @returns {boolean} `true` if Set contains at least one element,
 * `false` otherwise
 */
/**
 * Add an element to a Set. If a value is already there, a copy is not added.
 * @method
 * @instance
 * @name add
 * @memberOf Set
 * @param {Object} element
 */
/**
 * Remove an element from a Set. If no such element is there, nothing happens.
 * @method
 * @instance
 * @name delete
 * @memberOf Set
 * @param {Object} element
 */
/**
 * Remove all values from a Set.
 * @method
 * @instance
 * @name clear
 * @memberOf Set
 */

/**
 * @classdesc Common ancestor for all entity types.
 */
var BaseEntity = function () {
};

/**
 * Checks if a the entity has a field denoted by field name set, shows error message otherwise.
 *
 * @param {string} fieldName name of the field to check.
 * @param {string} message message to show in case the field is not set.
 */
BaseEntity.prototype.required = function (fieldName, message) {
};

/**
 * Checks that a value a field receives in the current transaction is equal to an expected one.
 *
 * @param {string} fieldName name of the field to check.
 * @param {string} expected expected value.
 *
 * @returns {boolean} `true` in case the expectation is met.
 */
BaseEntity.prototype.becomes = function (fieldName, expected) {
};

/**
 * Checks if a field is changed in the current transaction.
 *
 * @param {string} fieldName name of the field (e.g. 'summary') to check.
 *
 * @returns {boolean} `true` in case the field is changed in the current transaction.
 */
BaseEntity.prototype.isChanged = function (fieldName) {
};

/**
 * Returns the previous value of a single-valued field before an update was applied. In case the field is not changed
 * in the transaction, this value is equal to the current value of the field.
 *
 * @param {string} fieldName name of the field (e.g. 'updated') to get previous value of.
 *
 * @returns {Object} previous value of the field.
 */
BaseEntity.prototype.oldValue = function (fieldName) {
};

/**
 * Checks if a field can be read by a user.
 *
 * @param {string} fieldName name of the field.
 * @param {User} user to check access for.
 *
 * @returns {boolean} `true` in case the field can be read by the user.
 */
BaseEntity.prototype.canBeReadBy = function (fieldName, user) {
};

/**
 * Checks if a field can be written by a user.
 *
 * @param {string} fieldName name of the field.
 * @param {User} user to check access for.
 *
 * @returns {boolean} `true` in case the field can be written by the user.
 */
BaseEntity.prototype.canBeWrittenBy = function (fieldName, user) {
};



/**
 * @classdesc Stores settings for set of sprints.
 * @augments BaseEntity
 */
var Agile = function () {
  /**
   * The user who created the board.
   * @name Agile#author
   * @type {User}
   * @readonly
   */
  this.author = {/*this code is hidden*/};
  /**
   * Name of the board
   * @name Agile#name
   * @type {String}
   * @readonly
   */
  this.name = '';
  /**
   * The sprints created for the board.
   * @name Agile#sprints
   * @type {Set<Sprint>}
   * @readonly
   */
  this.sprints = {/*this code is hidden*/};
};
Agile.prototype = Object.create(BaseEntity.prototype);
Agile.prototype.constructor = BaseEntity;

/**
 * Finds a sprint of the board with a given name.
 * @param {String} name Sprint name.
 *
 * @returns {Sprint} Sprint or null.
 */
Agile.prototype.findSprintByName = function (name) {
  return {/*this code is hidden*/};
};

/**
 * Returns Agiles that have the specified name.
 * @static
 * @param {String} name The name to find by.
 *
 * @returns {Set<Agile>} Agiles found by the specified name.
 */
Agile.findByName = function (name) {
  return {/*this code is hidden*/};
};


/**
 * @classdesc Base class for issue work items.
 * @augments BaseEntity
 */
var BaseWorkItem = function () {
  /**
   * Work author.
   * @name BaseWorkItem#author
   * @type {User}
   * @readonly
   */
  this.author = {/*this code is hidden*/};
  /**
   * Date when the work item was created.
   * @name BaseWorkItem#created
   * @type {Date}
   * @readonly
   */
  this.created = new Date();
  /**
   * The User the workitem was created by.
   * @name BaseWorkItem#creator
   * @type {User}
   * @readonly
   */
  this.creator = {/*this code is hidden*/};
  /**
   * Work item description
   * @name BaseWorkItem#description
   * @type {String}
   */
  this.description = '';
  /**
   * Type of the work item.
   * @name BaseWorkItem#type
   * @type {WorkItemType}
   * @readonly
   */
  this.type = {/*this code is hidden*/};
  /**
   * Date when the work item was last updated.
   * @name BaseWorkItem#updated
   * @type {Date}
   * @readonly
   */
  this.updated = new Date();
};
BaseWorkItem.prototype = Object.create(BaseEntity.prototype);
BaseWorkItem.prototype.constructor = BaseEntity;


/**
 * @classdesc Represents a value of custom field of type build.
 * @augments Field
 */
var Build = function () {
  /**
   * Field type. To be used when defining rule requirements.
   * @name Build.fieldType
   * @type {String}
   * @readonly
   */
  this.fieldType = '';
  /**
   * The moment the build was assmbled.
   * @name Build#assembleDate
   * @type {Date}
   * @readonly
   */
  this.assembleDate = new Date();
};
Build.prototype = Object.create(Field.prototype);
Build.prototype.constructor = Field;


/**
 * @classdesc Represents a project custom field with a predefined set of values.
 * @augments ProjectCustomField
 */
var BundleProjectCustomField = function () {
  /**
   * Available values of this field.
   * @name BundleProjectCustomField#values
   * @type {Set<Field>}
   * @readonly
   */
  this.values = {/*this code is hidden*/};
};
BundleProjectCustomField.prototype = Object.create(ProjectCustomField.prototype);
BundleProjectCustomField.prototype.constructor = ProjectCustomField;

/**
 * Returns a value with the specified name for this field.
 * @param {String} name The name of the field value to search for.
 *
 * @returns {Field} Value found.
 */
BundleProjectCustomField.prototype.findValueByName = function (name) {
  return {/*this code is hidden*/};
};

/**
 * Returns a value with the specified position for this field.
 * @param {Number} ordinal The position of the field value to search by.
 *
 * @returns {Field} Value found.
 */
BundleProjectCustomField.prototype.findValueByOrdinal = function (ordinal) {
  return {/*this code is hidden*/};
};


/**
 * @classdesc Represents a value of a field with a predefined set of values.
 * @augments Field
 */
var EnumField = function () {
  /**
   * Field type. To be used when defining rule requirements.
   * @name EnumField.fieldType
   * @type {String}
   * @readonly
   */
  this.fieldType = '';
};
EnumField.prototype = Object.create(LocalizableField.prototype);
EnumField.prototype.constructor = LocalizableField;


/**
 * @classdesc Despite of the name of the class, it represents a value a field.
 * @augments BaseEntity
 */
var Field = function () {
  /**
   * Background color of the field as it is displayed in YouTrack UI.
   * @name Field#backgroundColor
   * @type {String}
   * @readonly
   */
  this.backgroundColor = '';
  /**
   * Foreground color of the field as it is displayed in YouTrack UI.
   * @name Field#foregroundColor
   * @type {String}
   * @readonly
   */
  this.foregroundColor = '';
  /**
   * String representation of the value.
   * @name Field#presentation
   * @type {String}
   * @readonly
   */
  this.presentation = '';
  /**
   * Date field type. To be used when defining rule requirements.
   * @name Field.dateType
   * @type {String}
   * @readonly
   */
  this.dateType = '';
  /**
   * Float field type. To be used when defining rule requirements.
   * @name Field.floatType
   * @type {String}
   * @readonly
   */
  this.floatType = '';
  /**
   * Integer field type. To be used when defining rule requirements.
   * @name Field.integerType
   * @type {String}
   * @readonly
   */
  this.integerType = '';
  /**
   * Period field type. To be used when defining rule requirements.
   * @name Field.periodType
   * @type {String}
   * @readonly
   */
  this.periodType = '';
  /**
   * String field type. To be used when defining rule requirements.
   * @name Field.stringType
   * @type {String}
   * @readonly
   */
  this.stringType = '';
  /**
   * Color index of the field.
   * @name Field#colorIndex
   * @type {Number}
   * @readonly
   */
  this.colorIndex = 0;
  /**
   * Description of the field, as visible is custom fields administration UI.
   * @name Field#description
   * @type {String}
   * @readonly
   */
  this.description = '';
  /**
   * Name of the field.
   * @name Field#name
   * @type {String}
   * @readonly
   */
  this.name = '';
  /**
   * Position of the field in list of all fields.
   * @name Field#ordinal
   * @type {Number}
   * @readonly
   */
  this.ordinal = 0;
};
Field.prototype = Object.create(BaseEntity.prototype);
Field.prototype.constructor = BaseEntity;


/**
 * @classdesc Field UserGroup type.
 * @augments ProjectCustomField
 */
var GroupProjectCustomField = function () {
  /**
   * Available values of this field.
   * @name GroupProjectCustomField#values
   * @type {Set<UserGroup>}
   * @readonly
   */
  this.values = {/*this code is hidden*/};
  /**
   * Default values.
   * @name GroupProjectCustomField#defaultValues
   * @type {Set<UserGroup>}
   */
  this.defaultValues = {/*this code is hidden*/};
};
GroupProjectCustomField.prototype = Object.create(ProjectCustomField.prototype);
GroupProjectCustomField.prototype.constructor = ProjectCustomField;

/**
 * Returns a value with the specified name for this field.
 * @param {String} name Name to search by.
 *
 * @returns {UserGroup} UserGroup that can be set to the field as a value and having the specified name.
 */
GroupProjectCustomField.prototype.findValueByName = function (name) {
  return {/*this code is hidden*/};
};


/**
 * @typedef {object} JsonForIssueConstructor
 * @property {User} reporter Issue reporter.
 * @property {Project} project Project that the new issue is to belong to.
 * @property {String} summary Issue summary.
 */
/**
 * @classdesc Represents YouTrack issue.
 * @augments BaseEntity
 *
 * @constructor Issue
 * @param {User|JsonForIssueConstructor} reporter Issue reporter. Alternatively, pass a JSON specified by {@link JsonForIssueConstructor}
 * @param {Project} [project] Project that the new issue is to belong to.
 * @param {String} [summary] Issue summary.
 */
var Issue = function (reporter, project, summary) {
  /**
   * `true` in case the issue becomes reported in the current transaction.
   * @example
   * if (issue.fields.Subsystem !== null && issue.fields.Assignee === null &&
   *     (((issue.isChanged('Subsystem') || issue.isChanged('project') && issue.isReported) ||
   *     issue.becomesReported) {
   *   issue.fields.Assignee = issue.fields.Subsystem.owner
   * }
   * @name Issue#becomesReported
   * @type {Boolean}
   * @readonly
   */
  this.becomesReported = true;
  /**
   * `true` in case the issue is assigned a state that is considered resolved in the current transaction.
   * @name Issue#becomesResolved
   * @type {Boolean}
   * @readonly
   */
  this.becomesResolved = true;
  /**
   * `true` in case the issue is assigned a state that is considered unresolved in the current transaction.
   * @name Issue#becomesUnresolved
   * @type {Boolean}
   * @readonly
   */
  this.becomesUnresolved = true;
  /**
   * The root issue in a tree of duplicates that are linked to the issue.
   * For instance, if `issueA` duplicates `issueB` and `issueB` duplicates
   * `issueC`, then `issueA.getDuplicateRoot()` would return `issueC`.
   * @name Issue#duplicateRoot
   * @type {Issue}
   * @readonly
   */
  this.duplicateRoot = {/*this code is hidden*/};
  /**
   * The comments that are edited in the current transaction.
   * Comments that are added and removed are not considered as edited.
   * These values are represented by the `issue.comments.added` and
   * `issue.comments.removed` properties.
   * @name Issue#editedComments
   * @type {Set<IssueComment>}
   * @readonly
   */
  this.editedComments = {/*this code is hidden*/};
  /**
   * Issue ID.
   * @example
   * user.notify('Issue is overdue', 'Please, look at the issue: ' + issue.id);
   * @name Issue#id
   * @type {String}
   * @readonly
   */
  this.id = '';
  /**
   * The issue URL.
   * @example
   * user.notify('Issue is overdue', 'Please, look at the issue: ' + issue.url);
   * @name Issue#url
   * @type {String}
   * @readonly
   */
  this.url = '';
  /**
   * Work items of the issue.
   * @name Issue#workItems
   * @type {Set<IssueWorkItem>}
   * @readonly
   */
  this.workItems = {/*this code is hidden*/};
  /**
   * `true` in case the issue is already reported or becomes reported in current transaction (stages 3 and 4). To apply changes to an issue draft (stage 2), use `!issue.isReported`.
   * @example
   * issue.links['depends on'].forEach(function(dep) {
   *   if (dep.isReported) {
   *     assert(dep.State.resolved, 'The issue has unresolved dependencies and thus cannot be set Fixed!');
   *   }
   * });
   * @name Issue#isReported
   * @type {Boolean}
   * @readonly
   */
  this.isReported = true;
  /**
   * `true` if the issue is assigned a state that is considered resolved.
   * @name Issue#isResolved
   * @type {Boolean}
   * @readonly
   */
  this.isResolved = true;
  /**
   * Whether any user has added the star tag to the issue.
   * @name Issue#isStarred
   * @type {Boolean}
   * @readonly
   */
  this.isStarred = true;
  /**
   * The list of attachments for the issue.
   * @name Issue#attachments
   * @type {Set<IssueAttachment>}
   * @readonly
   */
  this.attachments = {/*this code is hidden*/};
  /**
   * A list of comments for the issue.
   * @name Issue#comments
   * @type {Set<IssueComment>}
   * @readonly
   */
  this.comments = {/*this code is hidden*/};
  /**
   * The date when the issue was created.
   * @name Issue#created
   * @type {Date}
   * @readonly
   */
  this.created = new Date();
  /**
   * The text that is entered as the issue description.
   * @name Issue#description
   * @type {String}
   */
  this.description = '';
  /**
   * The issue number in the project.
   * @name Issue#numberInProject
   * @type {Number}
   * @readonly
   */
  this.numberInProject = 0;
  /**
   * The user group for which the issue is visible. If the property contains a null value, the issue is visible to the <b>All Users</b> group.
   * @name Issue#permittedGroup
   * @type {UserGroup}
   */
  this.permittedGroup = {/*this code is hidden*/};
  /**
   * User groups for which the issue is visible.
   * @name Issue#permittedGroups
   * @type {Set<UserGroup>}
   */
  this.permittedGroups = {/*this code is hidden*/};
  /**
   * Users for whom the issue is visible.
   * @name Issue#permittedUsers
   * @type {Set<User>}
   */
  this.permittedUsers = {/*this code is hidden*/};
  /**
   * The project to which the issue is assigned.
   * @name Issue#project
   * @type {Project}
   */
  this.project = {/*this code is hidden*/};
  /**
   * The user who reported the issue.
   * @example
   * issue.fields.Assignee = issue.reporter;
   * @name Issue#reporter
   * @type {User}
   * @readonly
   */
  this.reporter = {/*this code is hidden*/};
  /**
   * Issue resolution time.
   * @name Issue#resolved
   * @type {Date}
   * @readonly
   */
  this.resolved = new Date();
  /**
   * The text that is entered as the issue summary.
   * @name Issue#summary
   * @type {String}
   */
  this.summary = '';
  /**
   * The list of tags that are attached to an issue.
   * @name Issue#tags
   * @type {Set<IssueTag>}
   */
  this.tags = {/*this code is hidden*/};
  /**
   * The date when the issue was last updated.
   * @name Issue#updated
   * @type {Date}
   * @readonly
   */
  this.updated = new Date();
  /**
   * The user who last updated the issue.
   * @name Issue#updatedBy
   * @type {User}
   * @readonly
   */
  this.updatedBy = {/*this code is hidden*/};
  /**
   * The number of votes for an issue. For vote-related methods, see User#canVoteIssue, User#voteIssue, User#canUnvoteIssue, and User#unvoteIssue.
   * @name Issue#votes
   * @type {Number}
   * @readonly
   */
  this.votes = 0;

  /**
   * Checks if a the issue has a custom field set to non-empty value, shows error message otherwise.
   *
   * @method
   * @name required
   * @memberOf Fields
   *
   * @param {string} fieldName name of the field to check.
   * @param {string} message to show in case the field is not set.
   */
  /**
   * Checks that a value a custom field receives in the current transaction is equal to an expected one.
   *
   * @method
   * @name becomes
   * @memberOf Fields
   *
   * @param {string|ProjectCustomField} field the field to check.
   * @param {Object} expected expected value of the field.
   *
   * @returns {boolean} `true` in case the expectation is met.
   */
  /**
   * Checks if a custom field is changed in the current transaction.
   *
   * @method
   * @name isChanged
   * @memberOf Fields
   *
   * @param {string|ProjectCustomField} field the project custom field (e.g. 'State') or a reference to the field to check.
   *
   * @returns {boolean} `true` in case the custom field is changed in the current transaction.
   */
  /**
   * Returns the previous value of a single-valued custom field before an update was applied. In case the field is not changed
   * in the transaction, this value is equal to the current value of the field.
   *
   * @method
   * @name oldValue
   * @memberOf Fields
   *
   * @param {string|ProjectCustomField} field the project custom field (e.g. 'State') or a reference to get previous value of.
   *
   * @returns {Object} previous value of the custom field.
   */
  /**
   * Checks if a custom field can be read by a user.
   *
   * @method
   * @name canBeReadBy
   * @memberOf Fields
   *
   * @param {string|ProjectCustomField} field the project custom field to check read access to.
   * @param {User} user to check access for.
   *
   * @returns {boolean} `true` in case the custom field can be read by the user.
   */
  /**
   * Checks if a custom field can be written by a user.
   *
   * @method
   * @name canBeWrittenBy
   * @memberOf Fields
   *
   * @param {string|ProjectCustomField} field the project custom field to check write access to.
   * @param {User} user to check access for.
   *
   * @returns {boolean} `true` in case the custom field can be written by the user.
   */
  /**
   * @typedef {Object} Fields
   * @classdesc
   * Contains all issue custom fields.
   */
  /**
   * All issue custom fields. Note that custom fields are the fields like
   * `Assignee`, `State`, `Priority`, etc. (those values which are defined at
   * `Custom fields` administration UI).
   *
   * The values like `reporter`, `project`, etc., are called just "fields" and
   * accessed directly.
   *
   * @example
   * if (issue.fields.becomes(ctx.Priority, ctx.Priority.Critical) {
   *   issue.fields.Assignee = issue.project.leader;
   * }
   * @name Issue#fields
   * @type {Fields}
   */
  this.fields = {
    /**
     * Custom field of type Set<ProjectVersion>
     * @type {Set<ProjectVersion>}
     * @name Fields#['Affected versions']
     */
    'Affected versions': null,
    /**
     * Custom field of type User
     * @type {User}
     * @name Fields#Assignee
     */
    'Assignee': null,
    /**
     * Custom field of type Date
     * @type {Date}
     * @name Fields#['Due Date']
     */
    'Due Date': null,
    /**
     * Custom field of type Set<ProjectVersion>
     * @type {Set<ProjectVersion>}
     * @name Fields#['Fix versions']
     */
    'Fix versions': null,
    /**
     * Custom field of type Build
     * @type {Build}
     * @name Fields#['Fixed in build']
     */
    'Fixed in build': null,
    /**
     * Custom field of type EnumField
     * @type {EnumField}
     * @name Fields#Priority
     */
    'Priority': null,
    /**
     * Custom field of type State
     * @type {State}
     * @name Fields#State
     */
    'State': null,
    /**
     * Custom field of type OwnedField
     * @type {OwnedField}
     * @name Fields#Subsystem
     */
    'Subsystem': null,
    /**
     * Custom field of type EnumField
     * @type {EnumField}
     * @name Fields#Type
     */
    'Type': null,
    /**
     * Custom field of type Set<Sprint>
     * @type {Set<Sprint>}
     * @name Fields#sprints
     */
    'sprints': null
  };

  /**
   * Issue links (e.g. `relates to`, `parent for`, etc.). Each link is a
   * {@link Set} of {@link Issue} objects.
   * @example
   * if (issue.links['parent for'].added.isNotEmpty()) {
   *   issue.links['parent for'].added.forEach(function(subtask) {
   *     subtask.fields.Priority = issue.fields.Priority;
   *   });
   * }
   * @name Issue#links
   * @type {Object}
   */
  this.links = {
    'relates to': null,
    'depends on': null,
    'is required for': null,
    'duplicates': null,
    'is duplicated by': null,
    'subtask of': null,
    'parent for': null
  };
};
Issue.prototype = Object.create(BaseEntity.prototype);
Issue.prototype.constructor = BaseEntity;

/**
 * @typedef {object} JsonForIssueAddComment
 * @property {String} text The text to add to the issue as a comment.
 * @property {User} [author] The author of the comment.
 */
/**
 * Adds a comment to the issue.
 * Makes `issue.comments.isChanged` return `true` for the current transaction.
 * @param {String|JsonForIssueAddComment} text The text to add to the issue as a comment. Alternatively, pass a JSON specified by {@link JsonForIssueAddComment}
 * @param {User} [author] The author of the comment.
 *
 * @returns {IssueComment} A newly created comment.
 */
Issue.prototype.addComment = function (text, author) {
  return {/*this code is hidden*/};
};

/**
 * Add a tag with the specified name to an issue. YouTrack adds the first matching tag
 * that is visible to the current user. If a match is not found, a new private tag is created for the current user.
 * When you use this method to add the <em>star</em> tag to an issue, the current user is added to the list of watchers.
 * To add a tag that is not visible to the current user, use the applyCommand method instead. Use "add tag %tagName%" for the command parameter and specify the owner of the tag in the runAs parameter.
 * @example
 * addTag('doit');
 * @param {String} name The name of the tag to attach to the issue.
 *
 * @returns {IssueTag} Entity of type IssueTag representing a tag that has been added to the issue.
 */
Issue.prototype.addTag = function (name) {
  return {/*this code is hidden*/};
};

/**
 * @typedef {object} JsonForIssueAddWorkItem
 * @property {String} description either work item text.
 * @property {Set<Date>} [date] work date.
 * @property {User} [author] work author.
 * @property {Number} duration work duration in minutes.
 * @property {WorkItemType} [type] work type.
 */
/**
 * Adds a work item to the issue.
 * @param {String|JsonForIssueAddWorkItem} description either work item text. Alternatively, pass a JSON specified by {@link JsonForIssueAddWorkItem}
 * @param {Set<Date>} [date] work date.
 * @param {User} [author] work author.
 * @param {Number} [duration] work duration in minutes.
 * @param {WorkItemType} [type] work type.
 *
 * @returns {IssueWorkItem} a newly created work item.
 */
Issue.prototype.addWorkItem = function (description, date, author, duration, type) {
  return {/*this code is hidden*/};
};

/**
 * Applies a command to the issue.
 * @param {String} command The command that is applied to the issue.
 * @param {User} runAs Specifies the user by which the command is applied. If this parameter is not set, the command is applied on behalf of the current user.
 */
Issue.prototype.applyCommand = function (command, runAs) {};

/**
 * Removes all of the attachments from the issue.
 */
Issue.prototype.clearAttachments = function () {};

/**
 * Creates a copy of the issue.
 *
 * @returns {Issue} Copy of this issue.
 */
Issue.prototype.copy = function () {
  return {/*this code is hidden*/};
};

/**
 * Creates a copy of the issue. No workflows will be triggered for the newly created issue.
 * @Deprecated use copy() instead.
 *
 * @returns {Issue} copy of this issue.
 */
Issue.prototype.createCopy = function () {
  return {/*this code is hidden*/};
};

/**
 * Checks whether the specified tag is attached to an issue.
 * @param {String} name The name of the tag to check for the issue.
 *
 * @returns {Boolean} `true` in case the tag is attached.
 */
Issue.prototype.hasTag = function (name) {
  return true;
};

/**
 * Remove a tag with the specified name from an issue. If the specified tag is not attached to the issue, an exception is thrown.
 * This method first searches through tags owned by current user, then through all other tags visible.To remove a tag that is not visible to a current user, use the applyCommand method instead. Use "remove tag %tagName%"for the command parameter and specify the login for the owner of the tag in the runAs parameter.
 * @example
 * removeTag('waiting for reply');
 * @param {String} name The name of the tag to remove from the issue.
 *
 * @returns {IssueTag} Entity of type IssueTag representing a tag has been removed from the issue.
 */
Issue.prototype.removeTag = function (name) {
  return {/*this code is hidden*/};
};

/**
 * Converts text with wiki markup to HTML. Use this method to send "pretty" notifications.
 * @example
 * issue.Assignee.notify('Issue is commented', issue.wikify(comment.text));
 * @param {String} text The string of text to convert to HTML.
 *
 * @returns {String} Wikified string
 */
Issue.prototype.wikify = function (text) {
  return '';
};

/**
 * Creates a declaration of an action that a user can apply to one issue or multiple issues using YouTrack command
 * window or toolbar action.
 * Result of this call are normally exported under 'rule' property, otherwise it is not treated as a rule.
 * @example
 * var entities = require('@jetbrains/youtrack-scripting-api/entities');
 * exports.rule = entities.Issue.action({
 *   title: 'Log comments',
 *   command: 'log',
 *   guard: function(ctx) {
 *     return ctx.issue.isReported;
 *   },
 *   action: function(ctx) {
 *     ctx.issue.comments.forEach(function(comment) {
 *       console.log(comment.text);
 *     });
 *   }
 * });
 * @param {Object} ruleProperties JSON describing rule properties.
 * @param {string} ruleProperties.title human readable name of the rule. Will be displayed in YouTrack UI.
 * @param {string} ruleProperties.command custom command which will trigger this action.
 * @param {Issue~guardFunction} ruleProperties.guard function that is invoked to check if the action is applicable to an issue.
 * @param {Issue~actionFunction} ruleProperties.action function that is invoked when a user triggers this action.
 * @param {Requirements} ruleProperties.requirements that environment need to meet for this script to work.
 * @static
 *
 * @returns {Object} Rule representation.
 */
Issue.action = function (ruleProperties) {
  return {/*this code is hidden*/};
};

/**
 * Finds an issue by it's visible id.
 * @static
 * @param {String} id Issue id.
 *
 * @returns {Issue} The issue.
 */
Issue.findById = function (id) {
  return {/*this code is hidden*/};
};

/**
 * Creates a declaration of a rule that will be triggered when an issue is changed.
 * Result of this call are normally exported under 'rule' property, otherwise it is not treated as a rule.
 * @example
 * var entities = require('@jetbrains/youtrack-scripting-api/entities');
 * exports.rule = entities.Issue.onChange({
 *   title: 'On issue change, log its id',
 *   action: function(ctx) {
 *     console.log(ctx.issue.id);
 *   }
 * });
 * @param {Object} ruleProperties JSON describing rule properties.
 * @param {string} ruleProperties.title human readable name of the rule. Will be displayed in YouTrack UI.
 * @param {Issue~actionFunction} ruleProperties.action function that is invoked on issue change.
 * @param {Requirements} ruleProperties.requirements that environment need to meet for this script to work.
 * @static
 *
 * @returns {Object} Rule representation.
 */
Issue.onChange = function (ruleProperties) {
  return {/*this code is hidden*/};
};

/**
 * Creates a declaration of a rule that will be triggered on schedule.
 * Result of this call are normally exported under 'rule' property, otherwise it is not treated as a rule.
 * @example
 * var entities = require('@jetbrains/youtrack-scripting-api/entities');
 * exports.rule = entities.Issue.onSchedule({
 *   title: 'Log id of major issues every 5 seconds',
 *   search: '#Major',
 *   cron: '0/5 * * * * ?',
 *   action: function(ctx) {
 *     console.log(ctx.issue.id);
 *   }
 * });
 * @param {Object} ruleProperties JSON describing rule properties.
 * @param {string} [ruleProperties.title] human readable name of the rule. Will be displayed in YouTrack UI.
 * @param {string|function} ruleProperties.search YouTrack search string or a function with no parameters that returns such string.
 * The action will be execute for issues that match the search and belong to the project this rule is attached to.
 * @param {string} ruleProperties.cron cron expression that specifieds the triggering interval.
 * @param {Issue~actionFunction} ruleProperties.action function that is invoked on schedule for each issue that matches the search.
 * @param {Requirements} ruleProperties.requirements that environment need to meet for this script to work.
 * @static
 *
 * @returns {Object} Rule representation.
 */
Issue.onSchedule = function (ruleProperties) {
  return {/*this code is hidden*/};
};

/**
 * Creates declaration of a state machine. The state machine imposes restrictions on how an issue is transited from
 * one state to another. You can execute certain actions when the issue enters a state, exits it, or transits between
 * two distinct states.
 * Result of this call are normally exported under 'rule' property, otherwise it is not treated as a rule.
 * @example
 * var entities = require('@jetbrains/youtrack-scripting-api/entities');
 * exports.rule = entities.Issue.stateMachine({
 *   title: 'Status state-machine',
 *   fieldName: 'Status',
 *   states: {
 *     Open: {
 *       initial: true,
 *       transitions: {
 *         start: {
 *           targetState: 'In progress'
 *         }
 *       }
 *     },
 *     'In progress': {
 *       onEnter: function(ctx) {
 *         ctx.issue.fields.Assignee = ctx.currentUser;
 *       },
 *       transitions: {
 *         fix: {
 *           targetState: 'Fixed'
 *         },
 *         reopen: {
 *           targetState: 'Open'
 *         }
 *       }
 *     },
 *     Fixed: {
 *       transitions: {
 *       }
 *     }
 *   },
 *   requirements: {
 *     Assignee: {
 *       type: entities.User.fieldType
 *     }
 *   }
 * });
 * @param {Object} ruleProperties JSON describing rule properties.
 * @param {string} ruleProperties.title human readable name of the rule. Will be displayed in YouTrack UI.
 * @param {Object[]} ruleProperties.states list of field values and transitions between them.
 * @param {Requirements} ruleProperties.requirements that environment need to meet for this script to work.
 * @static
 *
 * @returns {Object} Rule representation.
 */
Issue.stateMachine = function (ruleProperties) {
  return {/*this code is hidden*/};
};


/**
 * @classdesc Represents an issue attachment.
 * @augments BaseEntity
 */
var IssueAttachment = function () {
  /**
   * The extension of a file the attachment represents.
   * @name IssueAttachment#extension
   * @type {String}
   * @readonly
   */
  this.extension = '';
  /**
   * The name of the attachment.
   * @name IssueAttachment#name
   * @type {String}
   * @readonly
   */
  this.name = '';
  /**
   * Size of the attachment in bytes.
   * @name IssueAttachment#size
   * @type {Number}
   * @readonly
   */
  this.size = 0;
  /**
   * `true` in case the attachment is removed.
   * @name IssueAttachment#isRemoved
   * @type {Boolean}
   * @readonly
   */
  this.isRemoved = true;
  /**
   * The user who created the attachment.
   * @name IssueAttachment#author
   * @type {User}
   * @readonly
   */
  this.author = {/*this code is hidden*/};
  /**
   * Time the attachment was created.
   * @name IssueAttachment#created
   * @type {Date}
   * @readonly
   */
  this.created = new Date();
  /**
   * Issue this attachment belongs to.
   * @name IssueAttachment#issue
   * @type {Issue}
   * @readonly
   */
  this.issue = {/*this code is hidden*/};
  /**
   * Dimentions of an image file. For an attachment representing an image, the value is
   * `rw=%width%&rh=%heigth%`, for a non-image file the value is `empty`.
   * @name IssueAttachment#metaData
   * @type {String}
   * @readonly
   */
  this.metaData = '';
  /**
   * A group whos members are allowed to access the attachment
   * @name IssueAttachment#permittedGroup
   * @type {UserGroup}
   */
  this.permittedGroup = {/*this code is hidden*/};
  /**
   * Groups whos members are allowed to access the attachment
   * @name IssueAttachment#permittedGroups
   * @type {Set<UserGroup>}
   */
  this.permittedGroups = {/*this code is hidden*/};
  /**
   * Users that are allowed to access the attachment
   * @name IssueAttachment#permittedUsers
   * @type {Set<User>}
   */
  this.permittedUsers = {/*this code is hidden*/};
  /**
   * Time the attachment was last updated.
   * @name IssueAttachment#updated
   * @type {Date}
   * @readonly
   */
  this.updated = new Date();
};
IssueAttachment.prototype = Object.create(BaseEntity.prototype);
IssueAttachment.prototype.constructor = BaseEntity;


/**
 * @classdesc Represents an issue comment.
 * @augments BaseEntity
 */
var IssueComment = function () {
  /**
   * The comment URL.
   * @example
   * user.notify('Somebody has written something', 'Have a look: ' + comment.url);
   * @name IssueComment#url
   * @type {String}
   * @readonly
   */
  this.url = '';
  /**
   * The user who created the comment.
   * @name IssueComment#author
   * @type {User}
   * @readonly
   */
  this.author = {/*this code is hidden*/};
  /**
   * Time the comment was created.
   * @name IssueComment#created
   * @type {Date}
   * @readonly
   */
  this.created = new Date();
  /**
   * The issue the comment belongs to.
   * @name IssueComment#issue
   * @type {Issue}
   * @readonly
   */
  this.issue = {/*this code is hidden*/};
  /**
   * A group whos members are allowed to access the comment.
   * @name IssueComment#permittedGroup
   * @type {UserGroup}
   */
  this.permittedGroup = {/*this code is hidden*/};
  /**
   * Groups whos members are allowed to access the comment.
   * @name IssueComment#permittedGroups
   * @type {Set<UserGroup>}
   */
  this.permittedGroups = {/*this code is hidden*/};
  /**
   * Users that are allowed to access the comment.
   * @name IssueComment#permittedUsers
   * @type {Set<User>}
   */
  this.permittedUsers = {/*this code is hidden*/};
  /**
   * The text of the comment.
   * @name IssueComment#text
   * @type {String}
   */
  this.text = '';
  /**
   * Time the comment was last updated.
   * @name IssueComment#updated
   * @type {Date}
   * @readonly
   */
  this.updated = new Date();
  /**
   * The user the comment was last updated by.
   * @name IssueComment#updatedBy
   * @type {User}
   * @readonly
   */
  this.updatedBy = {/*this code is hidden*/};
};
IssueComment.prototype = Object.create(BaseEntity.prototype);
IssueComment.prototype.constructor = BaseEntity;


/**
 * @classdesc Represets a tag.
 * @augments WatchFolder
 */
var IssueTag = function () {
  /**
   * The User who created the tag.
   * @name IssueTag#owner
   * @type {User}
   * @readonly
   */
  this.owner = {/*this code is hidden*/};
};
IssueTag.prototype = Object.create(WatchFolder.prototype);
IssueTag.prototype.constructor = WatchFolder;

/**
 * Get all tags with this name, which current user can see.
 * Own user tags with this name go first.
 * @static
 * @param {String} name tag name
 *
 * @returns {Set<IssueTag>} sequence of IssueTags
 */
IssueTag.findByName = function (name) {
  return {/*this code is hidden*/};
};


/**
 * @classdesc Work item, added to issue
 * @augments BaseWorkItem
 */
var IssueWorkItem = function () {
  /**
   * Day, when work was done. This property contains timestamp of midnight in UTC for the day, when work was done.
   * @name IssueWorkItem#date
   * @type {Date}
   * @readonly
   */
  this.date = new Date();
  /**
   * Work duration in minutes
   * @name IssueWorkItem#duration
   * @type {Number}
   * @readonly
   */
  this.duration = 0;
};
IssueWorkItem.prototype = Object.create(BaseWorkItem.prototype);
IssueWorkItem.prototype.constructor = BaseWorkItem;


/**
 * @classdesc Represents a value that has a user associated with it, a so called owner.
 * @augments Field
 */
var OwnedField = function () {
  /**
   * Field type. To be used when defining rule requirements.
   * @name OwnedField.fieldType
   * @type {String}
   * @readonly
   */
  this.fieldType = '';
  /**
   * The associated user.
   * @name OwnedField#owner
   * @type {User}
   * @readonly
   */
  this.owner = {/*this code is hidden*/};
};
OwnedField.prototype = Object.create(Field.prototype);
OwnedField.prototype.constructor = Field;


/**
 * @classdesc Project custom field, that stores period value
 * @augments SimpleProjectCustomField
 */
var PeriodProjectCustomField = function () {
};
PeriodProjectCustomField.prototype = Object.create(SimpleProjectCustomField.prototype);
PeriodProjectCustomField.prototype.constructor = SimpleProjectCustomField;


/**
 * @classdesc Represents a YouTrack project.
 * @augments BaseEntity
 */
var Project = function () {
  /**
   * The key of the project. Use instead of project.shortName, which is deprecated.
   * @name Project#key
   * @type {String}
   * @readonly
   */
  this.key = '';
  /**
   * The name of the project.
   * @name Project#name
   * @type {String}
   * @readonly
   */
  this.name = '';
  /**
   * The email address that is used to send notifications for the project.
   * If a From address is not set for the project, the From address for the YouTrack server is returned.
   * @example
   * if (issue.becomesReported) {
   *   var lastRelatedEmails = issue.fields['Last message related emails'];
   *   if (lastRelatedEmails) {
   *     lastRelatedEmails.split(' ').forEach(function (email) {
   *       if (email && email.equalsIgnoreCase(issue.project.notificationEmail)) {
   *         var allRelatedEmails = issue.fields['All related emails'];
   *         if (!allRelatedEmails) {
   *           issue.fields['All related emails'] = email;
   *         } else if (!(allRelatedEmails.split(' ').has(email))) {
   *           issue.fields['All related emails'] = allRelatedEmails + ' ' + email;
   *         }
   *       }
   *     });
   *     issue.fields['Last message related emails'] = null;
   *   }
   * }
   * @name Project#notificationEmail
   * @type {String}
   * @readonly
   */
  this.notificationEmail = '';
  /**
   * `true` in case the project is archived.
   * @name Project#isArchived
   * @type {Boolean}
   * @readonly
   */
  this.isArchived = true;
  /**
   * The description of the project. Available in project administration UI.
   * @name Project#description
   * @type {String}
   * @readonly
   */
  this.description = '';
  /**
   * Custom fields configured for the project.
   * @name Project#fields
   * @type {Set<ProjectCustomField>}
   * @readonly
   */
  this.fields = {/*this code is hidden*/};
  /**
   * All issues of the project.
   * @name Project#issues
   * @type {Set<Issue>}
   * @readonly
   */
  this.issues = {/*this code is hidden*/};
  /**
   * The leader of the project. Available in project administration UI.
   * @name Project#leader
   * @type {User}
   * @readonly
   */
  this.leader = {/*this code is hidden*/};
};
Project.prototype = Object.create(BaseEntity.prototype);
Project.prototype.constructor = BaseEntity;

/**
 * Returns the project's field with the specified name.
 * @param {String} name Field name.
 *
 * @returns {ProjectCustomField} Field with the specified name.
 */
Project.prototype.findFieldByName = function (name) {
  return {/*this code is hidden*/};
};

/**
 * Finds a project by short name (id).
 * @static
 * @param {String} key The key to find by.
 *
 * @returns {Project} Project, or null in case no project exists with the provided short name.
 */
Project.findByKey = function (key) {
  return {/*this code is hidden*/};
};

/**
 * Finds a project by name.
 * @static
 * @param {String} name name to find by.
 *
 * @returns {Project} project, or null in case no project exists with the provided name.
 */
Project.findByName = function (name) {
  return {/*this code is hidden*/};
};


/**
 * @classdesc Represents a custom field attached to a Project.
 * @augments BaseEntity
 */
var ProjectCustomField = function () {
  /**
   * Localized name of the field.
   * @name ProjectCustomField#localizedName
   * @type {String}
   * @readonly
   */
  this.localizedName = '';
  /**
   * Name of the field.
   * @name ProjectCustomField#name
   * @type {String}
   * @readonly
   */
  this.name = '';
  /**
   * The text to be displayed for this field in case it has not value.
   * @name ProjectCustomField#nullValueText
   * @type {String}
   * @readonly
   */
  this.nullValueText = '';
};
ProjectCustomField.prototype = Object.create(BaseEntity.prototype);
ProjectCustomField.prototype.constructor = BaseEntity;

/**
 * Returns the background color that is used for this field value in the specified issue.
 * Can return `null`, `"white"` or a hex color presentation.
 * @param {Issue} issue The issue for which the background color is returned.
 *
 * @returns {String} The background color that is used for this field value in the specified issue.
 */
ProjectCustomField.prototype.getBackgroundColor = function (issue) {
  return '';
};

/**
 * Returns the foreground color that is used for this field value in the specified issue.
 * Can return `null`, `"white"` or a hex color presentation.
 * @param {Issue} issue The issue for which the foreground color is returned.
 *
 * @returns {String} The foreground color that is used for this field value in the specified issue.
 */
ProjectCustomField.prototype.getForegroundColor = function (issue) {
  return '';
};

/**
 * Returns the string presentation of the value that is used for this field in the specified issue.
 * @param {Issue} issue The issue for which the value presentation is returned.
 *
 * @returns {String} The string presentation of the value.
 */
ProjectCustomField.prototype.getValuePresentation = function (issue) {
  return '';
};


/**
 * @classdesc Represents a version field value.
 * @augments Field
 */
var ProjectVersion = function () {
  /**
   * `true` in case the version is archived.
   * @name ProjectVersion#isArchived
   * @type {Boolean}
   * @readonly
   */
  this.isArchived = true;
  /**
   * `true` in case the version is released.
   * @name ProjectVersion#isReleased
   * @type {Boolean}
   * @readonly
   */
  this.isReleased = true;
  /**
   * Field type. To be used when defining rule requirements.
   * @name ProjectVersion.fieldType
   * @type {String}
   * @readonly
   */
  this.fieldType = '';
  /**
   * Release date of the version.
   * @name ProjectVersion#releaseDate
   * @type {Date}
   * @readonly
   */
  this.releaseDate = new Date();
};
ProjectVersion.prototype = Object.create(Field.prototype);
ProjectVersion.prototype.constructor = Field;


/**
 * @classdesc Represents a saved query (search).
 * @augments WatchFolder
 */
var SavedQuery = function () {
  /**
   * The User who created the saved query (search).
   * @name SavedQuery#owner
   * @type {User}
   * @readonly
   */
  this.owner = {/*this code is hidden*/};
  /**
   * The query string.
   * @name SavedQuery#query
   * @type {String}
   * @readonly
   */
  this.query = '';
};
SavedQuery.prototype = Object.create(WatchFolder.prototype);
SavedQuery.prototype.constructor = WatchFolder;

/**
 * Get saved queries (searches) visible to current user with a given name.
 * User's own saved queries go first.
 * @static
 * @param {String} name saved query name
 *
 * @returns {Set<SavedQuery>} sequence of SavedQuery
 */
SavedQuery.findByName = function (name) {
  return {/*this code is hidden*/};
};


/**
 * @classdesc Base class for fields, that store "simple" values (like strings and numbers)
 * @augments ProjectCustomField
 */
var SimpleProjectCustomField = function () {
};
SimpleProjectCustomField.prototype = Object.create(ProjectCustomField.prototype);
SimpleProjectCustomField.prototype.constructor = ProjectCustomField;


/**
 * @classdesc Every sprint may contains issues from one project or from several projects.
 * Sprint is a agile wrapper for project version
 * If project version is shared between several projects, such sprint is a cross-project sprint.
 * Sprint uses states from states bundle of associated agile.
 * Content of particular sprint column (columns is associated with state) is:
 *   fix for: version state: column
 * @augments BaseEntity
 */
var Sprint = function () {
  /**
   * The name of the sprint.
   * @name Sprint#name
   * @type {String}
   * @readonly
   */
  this.name = '';
  /**
   * `true` in case the sprint is archived.
   * @name Sprint#isArchived
   * @type {Boolean}
   * @readonly
   */
  this.isArchived = true;
  /**
   * Agile board the sprint belongs to.
   * @name Sprint#agile
   * @type {Agile}
   * @readonly
   */
  this.agile = {/*this code is hidden*/};
  /**
   * End date of the sprint.
   * @name Sprint#finish
   * @type {Date}
   * @readonly
   */
  this.finish = new Date();
  /**
   * Start date of the sprint.
   * @name Sprint#start
   * @type {Date}
   * @readonly
   */
  this.start = new Date();
};
Sprint.prototype = Object.create(BaseEntity.prototype);
Sprint.prototype.constructor = BaseEntity;

/**
 * Checks if the specified issue is a swimlane of the board the sprint belongs to.
 * @param {Issue} issue Issue to check.
 *
 * @returns {Boolean} `true` in case the issue is a swimlane on the sprint's board.
 */
Sprint.prototype.isSwimlane = function (issue) {
  return true;
};


/**
 * @classdesc Represents an issue state, i.e. a value of a state field.
 * @augments Field
 */
var State = function () {
  /**
   * Field type. To be used when defining rule requirements.
   * @name State.fieldType
   * @type {String}
   * @readonly
   */
  this.fieldType = '';
  /**
   * `true` in case an issue in this state is considered resolved.
   * @name State#isResolved
   * @type {Boolean}
   * @readonly
   */
  this.isResolved = true;
};
State.prototype = Object.create(LocalizableField.prototype);
State.prototype.constructor = LocalizableField;


/**
 * @classdesc Represents YouTrack user.
 * @augments BaseEntity
 */
var User = function () {
  /**
   * The email address of the user.
   * @name User#email
   * @type {String}
   * @readonly
   */
  this.email = '';
  /**
   * The full name of the user as seen in their profile.
   * @name User#fullName
   * @type {String}
   * @readonly
   */
  this.fullName = '';
  /**
   * The login of the user.
   * @name User#login
   * @type {String}
   * @readonly
   */
  this.login = '';
  /**
   * User's time zone id.
   * @name User#timeZoneId
   * @type {String}
   * @readonly
   */
  this.timeZoneId = '';
  /**
   * The full name of the user. If the full name is not set, the login is returned.
   * @name User#visibleName
   * @type {String}
   * @readonly
   */
  this.visibleName = '';
  /**
   * `true` in case the user account is banned.
   * @name User#isBanned
   * @type {Boolean}
   * @readonly
   */
  this.isBanned = true;
  /**
   * Current (logged in user).
   * @name User.current
   * @type {User}
   * @readonly
   */
  this.current = {/*this code is hidden*/};
  /**
   * Field type. To be used when defining rule requirements.
   * @name User.fieldType
   * @type {String}
   * @readonly
   */
  this.fieldType = '';
};
User.prototype = Object.create(BaseEntity.prototype);
User.prototype.constructor = BaseEntity;

/**
 * Checks whether the user is able to remove their vote from the specified issue.
 * @param {Issue} issue The issue to check.
 *
 * @returns {Boolean} `true` in case the user can unvote.
 */
User.prototype.canUnvoteIssue = function (issue) {
  return true;
};

/**
 * Checks whether the user is able to vote for the specified issue.
 * @param {Issue} issue The issue to check.
 *
 * @returns {Boolean} `true` in case the user can vote.
 */
User.prototype.canVoteIssue = function (issue) {
  return true;
};

/**
 * Returns a tag with the specified name that is shared with but not owned by the user. If such a tag does not exist, a null value is returned.
 * @param {String} name The name of the tag.
 *
 * @returns {IssueTag} The tag.
 */
User.prototype.getSharedTag = function (name) {
  return {/*this code is hidden*/};
};

/**
 * Returns a tag that is visible to the user.
 * @param {String} name The name of the tag.
 * @param {Boolean} createIfNotExists If `true` and the specified tag does not exist or is not visible to the user and the user has permission to create tags, a new tag with the specified name is created.
 *
 * @returns {IssueTag} The tag.
 */
User.prototype.getTag = function (name, createIfNotExists) {
  return {/*this code is hidden*/};
};

/**
 * Checks whether the user has the specified role in the specified project. In case no project is specified, checks if the user has the specified role in any project.
 * @param {String} roleName The name of the role to check for.
 * @param {Project} project project to check access for.
 *
 * @returns {Boolean} `true` in case the user has the specified role.
 */
User.prototype.hasRole = function (roleName, project) {
  return true;
};

/**
 * Checks whether the user is a member of the specified group.
 * @param {String} groupName The name of the group to check for.
 *
 * @returns {Boolean} `true` in case the user is a member of the specified group.
 */
User.prototype.isInGroup = function (groupName) {
  return true;
};

/**
 * @typedef {object} JsonForUserNotify
 * @property {String} subject The subject line of the email notification.
 * @property {String} body The message text of the email notification.
 * @property {Boolean} [ignoreNotifyOnOwnChangesSetting] If `false`, the message is not sent when changes are performed on behalf of the current user. The corresponding flag is unchecked. Otherwise, the message is sent anyway.
 * @property {Project} [project] The email address configured for the project will be used as 'from'.
 */
/**
 * Sends an email notification to the email address that is set in the user profile.
 * @param {String|JsonForUserNotify} subject The subject line of the email notification. Alternatively, pass a JSON specified by {@link JsonForUserNotify}
 * @param {String} [body] The message text of the email notification.
 * @param {Boolean} [ignoreNotifyOnOwnChangesSetting] If `false`, the message is not sent when changes are performed on behalf of the current user. The corresponding flag is unchecked. Otherwise, the message is sent anyway.
 * @param {Project} [project] The email address configured for the project will be used as 'from'.
 */
User.prototype.notify = function (subject, body, ignoreNotifyOnOwnChangesSetting, project) {};

/**
 * Sends a notification message over Jabber. Similar to `notify`
 * method, the message won't be sent on own changes and corresponding flag unchecked.
 * @param {String} message The message text for the Jabber notification.
 */
User.prototype.sendJabber = function (message) {};

/**
 * An alias for notify(subject, body, true)
 * @param {String} subject see notify for the descripton
 * @param {String} body see notify for the description
 */
User.prototype.sendMail = function (subject, body) {};

/**
 * Removes a vote on behalf of the user from the issue, if allowed.
 * @param {Issue} issue The issue from which the vote is removed.
 */
User.prototype.unvoteIssue = function (issue) {};

/**
 * Removes the current user from the list of watchers for the issue
 * (remove `Star` tag).
 * @param {Issue} issue The issue to from which the user is removed as a watcher.
 */
User.prototype.unwatchIssue = function (issue) {};

/**
 * Adds a vote on behalf of the user to the issue, if allowed.
 * @param {Issue} issue The issue to which the vote is added.
 */
User.prototype.voteIssue = function (issue) {};

/**
 * Adds the current user to the issue as a watcher (add `Star` tag).
 * @param {Issue} issue The issue to which the user is added as a watcher.
 */
User.prototype.watchIssue = function (issue) {};

/**
 * Finds a user by login.
 * @static
 * @param {String} login login to find by.
 *
 * @returns {User} user, or null in case no user exists with the provided login.
 */
User.findByLogin = function (login) {
  return {/*this code is hidden*/};
};


/**
 * @classdesc Represents a user group.
 * @augments BaseEntity
 */
var UserGroup = function () {
  /**
   * All of the users who are members of the group.
   * @name UserGroup#users
   * @type {Set<User>}
   * @readonly
   */
  this.users = {/*this code is hidden*/};
  /**
   * `true` in case the group is All Users group.
   * @name UserGroup#isAllUsersGroup
   * @type {Boolean}
   * @readonly
   */
  this.isAllUsersGroup = true;
  /**
   * `true` in case newly registered users are automatically added to this group.
   * @name UserGroup#isAutoJoin
   * @type {Boolean}
   * @readonly
   */
  this.isAutoJoin = true;
  /**
   * All Users group.
   * @name UserGroup.allUsersGroup
   * @type {UserGroup}
   * @readonly
   */
  this.allUsersGroup = {/*this code is hidden*/};
  /**
   * Field type. To be used when defining rule requirements.
   * @name UserGroup.fieldType
   * @type {String}
   * @readonly
   */
  this.fieldType = '';
  /**
   * The description of the group.
   * @name UserGroup#description
   * @type {String}
   * @readonly
   */
  this.description = '';
  /**
   * The name of the group.
   * @name UserGroup#name
   * @type {String}
   * @readonly
   */
  this.name = '';
};
UserGroup.prototype = Object.create(BaseEntity.prototype);
UserGroup.prototype.constructor = BaseEntity;

/**
 * Sends an email notification to all of the users who are members of the group.
 * @example
 * issue.oldValue('permittedGroup').notifyAllUsers('Visibility has been changed',
 *   'The visibility group for the issue ' + issue.getId() +
 *   ' has been changed to ' + permittedGroup.name);
 * @param {String} subject The subject line of the email notification.
 * @param {String} body The message text of the email notification.
 */
UserGroup.prototype.notifyAllUsers = function (subject, body) {};

/**
 * Finds a group by name.
 * @static
 * @param {String} name name to find by.
 *
 * @returns {UserGroup} user group, or null in case no group exists with the provided name.
 */
UserGroup.findByName = function (name) {
  return {/*this code is hidden*/};
};


/**
 * @classdesc Represents a field that has s as values and taht is attached to a Project.
 * @augments ProjectCustomField
 */
var UserProjectCustomField = function () {
  /**
   * Available values of this field.
   * @name UserProjectCustomField#values
   * @type {Set<User>}
   * @readonly
   */
  this.values = {/*this code is hidden*/};
  /**
   * Dafault value of this field.
   * @name UserProjectCustomField#defaultUsers
   * @type {Set<User>}
   * @readonly
   */
  this.defaultUsers = {/*this code is hidden*/};
};
UserProjectCustomField.prototype = Object.create(ProjectCustomField.prototype);
UserProjectCustomField.prototype.constructor = ProjectCustomField;

/**
 * Returns a value with the specified login for this field.
 * @param {String} login Login to search by.
 *
 * @returns {User} User that can be set to the field as a value and having the specified login.
 */
UserProjectCustomField.prototype.findValueByLogin = function (login) {
  return {/*this code is hidden*/};
};


/**
 * @classdesc Represents a common ancestor of classes representing tag and saved query (search).
 * @augments BaseEntity
 */
var WatchFolder = function () {
  /**
   * The name of the tag or the saved query (search).
   * @name WatchFolder#name
   * @type {String}
   * @readonly
   */
  this.name = '';
  /**
   * The `UserGroup` for whom the tag or the saved query (search) is visible.
   * If the tag or the saved query (search) is visible only to its owner, returns `null`.
   * @name WatchFolder#shareGroup
   * @type {UserGroup}
   * @readonly
   */
  this.shareGroup = {/*this code is hidden*/};
  /**
   * The `UserGroup` for whom the tag or the saved query (search) is updatable.
   * If the tag or the saved query (search) is updatable only to its owner, returns `null`.
   * @name WatchFolder#updateShareGroup
   * @type {UserGroup}
   * @readonly
   */
  this.updateShareGroup = {/*this code is hidden*/};
};
WatchFolder.prototype = Object.create(BaseEntity.prototype);
WatchFolder.prototype.constructor = BaseEntity;


/**
 * @classdesc Represents work type. Examples: Development, Testing.
 * @augments BaseEntity
 */
var WorkItemType = function () {
  /**
   * Name of the work item type.
   * @name WorkItemType#name
   * @type {String}
   * @readonly
   */
  this.name = '';
};
WorkItemType.prototype = Object.create(BaseEntity.prototype);
WorkItemType.prototype.constructor = BaseEntity;

/**
 * Returns work item types configured for the project.
 * @static
 * @param {Project} project will be used to look for work item types.
 *
 * @returns {Set<WorkItemType>} work item types.
 */
WorkItemType.findByProject = function (project) {
  return {/*this code is hidden*/};
};


exports.Agile = Agile;
exports.BaseWorkItem = BaseWorkItem;
exports.Build = Build;
exports.BundleProjectCustomField = BundleProjectCustomField;
exports.EnumField = EnumField;
exports.Field = Field;
exports.GroupProjectCustomField = GroupProjectCustomField;
exports.Issue = Issue;
exports.IssueAttachment = IssueAttachment;
exports.IssueComment = IssueComment;
exports.IssueTag = IssueTag;
exports.IssueWorkItem = IssueWorkItem;
exports.OwnedField = OwnedField;
exports.PeriodProjectCustomField = PeriodProjectCustomField;
exports.Project = Project;
exports.ProjectCustomField = ProjectCustomField;
exports.ProjectVersion = ProjectVersion;
exports.SavedQuery = SavedQuery;
exports.SimpleProjectCustomField = SimpleProjectCustomField;
exports.Sprint = Sprint;
exports.State = State;
exports.User = User;
exports.UserGroup = UserGroup;
exports.UserProjectCustomField = UserProjectCustomField;
exports.WatchFolder = WatchFolder;
exports.WorkItemType = WorkItemType;
